#!/usr/bin/python3
# -*- coding: utf-8 -*-

"""
This module contains everything related too buttons
"""

# pylint: disable=too-few-public-methods


class MouseButtons:
    """
    https://tronche.com/gui/x/xlib/events/keyboard-pointer/keyboard-pointer.html\n
    /usr/include/X11/X.h: 259-263
    """
    LEFT = 1
    RIGHT = 2
    MIDDLE = 3
    FORWARD = 4
    BACKWARD = 5

# pylint: enable=too-few-public-methods
